#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include "select.h"
#include "telnet.h"
#include "ftp.h"
#include "sys.h"
#include "err.h"
#include "conn.h"
#include "scan.h"
#include "log.h"
#include "conf-home.h"

char main_inbuf[128];
#define inbuf main_inbuf

static void mainloop()
{
 int nfds;
 fd_set rfds;
 fd_set wfds;
 int r;

 for (;;)
  {
   FD_ZERO(&rfds);
   FD_ZERO(&wfds);

   FD_SET(0,&rfds);
   nfds = 1;
   ftp_fdrwatch(&nfds,&rfds);
   ftp_fdwwatch(&nfds,&wfds);

   do
     r = select(nfds,&rfds,&wfds,(fd_set *) 0,(struct timeval *) 0);
   while ((r == -1) && (errno == EINTR));
   if (r == -1) break;

   if (FD_ISSET(0,&rfds))
    {
     r = sys_read(0,inbuf,sizeof(inbuf));
     if (r <= 0)
       break;
     telnet_feed(inbuf,r);
    }
   ftp_fdrcheck(&rfds);
   ftp_fdwcheck(&wfds);
  }
}

void sigalrm()
{
 _exit(1);
}

void main(argc,argv)
int argc;
char **argv;
{
 unsigned long uid;
 unsigned long gid;
 unsigned int len;

 if (!argv[1]) goto death;
 if (!argv[2]) goto death;
 len = scan_ulong(argv[1],&uid);
 if (!len || argv[1][len]) goto death;
 len = scan_ulong(argv[2],&gid);
 if (!len || argv[2][len]) goto death;

 if (chdir(CONF_HOME) == -1) goto death;
 log_fd = open("log",O_WRONLY | O_APPEND | O_CREAT,0644);
 if (chdir("pub") == -1) goto death;
 if (chroot(".") == -1) goto death;
 if (setgid((gid_t) gid) == -1) goto death;
 if (setuid((uid_t) uid) == -1) goto death;

 signal(SIGPIPE,SIG_IGN);
 signal(SIGALRM,sigalrm);
 alarm(86400);

 if (ftp_init() == -1) goto death;
 if (conn_undostupidity(0) == -1) goto death;
 mainloop();
 ftp_quit((char *) 0);

 death:
 err_init();
 _exit(1);
}
