#include <errno.h>
extern int errno;
#include "ftp.h"
#include "conn.h"
#include "out.h"
#include "wd.h"
#include "get.h"
#include "err.h"
#include "log.h"
#include "sys.h"
#include "scan6.h"
#include "fmt6.h"
#include "select.h"

static struct conn_addr usc;
static struct conn_addr themc;
static struct conn_addr usd; /* always meaningful */
static struct conn_addr themd; /* meaningful if flagactive or flagaccepted */

static int fdlisten = -1;
/* -1 if flagactive or if !flagactive && flagaccepted */
static int fddata = -1;
/* -1 if flagactive && !flagwriting or if !flagactive && !flagaccepted */

static int flagactive = 1; /* 0 after PASV; 1 after PORT */
static int flagaccepted = 0; /* must be 0 if flagactive or if !flagwriting */
static int flagwriting = 0; /* 1 after RETR/LIST/... */
static int flagcorete = 0;
static int flagascii = 1;

static int transferlen; /* defined if flagwriting */
static int transferpos; /* defined if flagwriting */
char ftp_buf[512];
#define buf ftp_buf

static void usd_default()
{
 usd = usc;
 usd.p1 = usd.p2 = 0;
}

static void endtransfer(e)
int e;
{
 int saveerrno;
 saveerrno = errno;
 /* flagwriting is 1 */
 get_close();
 if (flagactive || flagaccepted) { close(fddata); fddata = -1; }
 if (!flagactive && !flagaccepted) { close(fdlisten); fdlisten = -1; }
 flagwriting = 0;
 flagactive = 1;
 flagaccepted = 0;
 usd_default();
 switch(e)
  {
   case -1:
     err_local();
     break;
   case 0:
     buf[fmttrans1(buf,transferpos)] = 0;
     out(buf);
     break;
   case 1:
     err_conn(saveerrno);
     break;
   case 2:
     err_closing();
     break;
  }
}

static void datawritable()
{
 int w;
 char ch;
 int len;
 char *x;
 if (!flagcorete)
  {
   flagcorete = 1;
   if (sys_read(fddata,&ch,0) == -1) 
#ifdef EAGAIN
if (errno != EAGAIN) /* Solaris dorks */
#endif
    { endtransfer(1); return; }
  }
 x = get_more(&len);
 if (!x) { endtransfer(len); return; } /* XXX: interface is sloppy */
 w = sys_write(fddata,x,len);
 if (w <= 0) { endtransfer(1); return; }
 get_done(w);
 transferpos += w;
}

void ftp_fdrwatch(nfds,rfds)
int *nfds;
fd_set *rfds;
{
 /* note: we don't accept connections until we start a transfer */
 if (!flagactive && !flagaccepted && flagwriting)
  {
   FD_SET(fdlisten,rfds);
   if (fdlisten >= *nfds) *nfds = fdlisten + 1;
  }
}

void ftp_fdrcheck(rfds)
fd_set *rfds;
{
 if (!flagactive && !flagaccepted && flagwriting)
   if (FD_ISSET(fdlisten,rfds))
    {
     fddata = conn_accept(fdlisten,&themd);
     if (fddata == -1)
      {
       err_trouble();
       flagwriting = 0;
      }
     else
      {
       close(fdlisten); fdlisten = -1;
       flagaccepted = 1;
      }
    }
}

void ftp_fdwwatch(nfds,wfds)
int *nfds;
fd_set *wfds;
{
 if (flagwriting && (flagactive || flagaccepted))
  {
   FD_SET(fddata,wfds);
   if (fddata >= *nfds) *nfds = fddata + 1;
  }
}

void ftp_fdwcheck(wfds)
fd_set *wfds;
{
 if (flagwriting && (flagactive || flagaccepted))
   if (FD_ISSET(fddata,wfds))
     datawritable();
}

static void passive_off()
{
 /* flagwriting is 0 */
 if (!flagactive)
  {
   close(fdlisten); fdlisten = -1;
   flagactive = 1;
   usd_default();
  }
}

static int passive_on()
{
 /* flagwriting is 0 */
 passive_off();
 fdlisten = conn_server(&usd);
 if (fdlisten == -1)
   return -1;
 flagactive = 0;
 return 0;
}

int ftp_init()
{
 if (conn_us(0,&usc) == -1) return -1;
 usd_default();
 if (conn_them(1,&themc) == -1) return -1;
 themd = themc;
 err_hi();
 log_init(&usc,&themc);
 return 0;
}

void ftp_quit(arg) char *arg;
{
 if (flagwriting) { err_transferring(); return; }
 err_bye();
 _exit(0);
}
void ftp_stru(arg) char *arg;
{
 if (arg && ((*arg == 'f') || (*arg == 'F')))
   err_struok();
 else
   err_strubad();
}
void ftp_mode(arg) char *arg;
{
 if (arg && ((*arg == 's') || (*arg == 'S')))
   err_modeok();
 else
   err_modebad();
}
void ftp_type(arg) char *arg;
{
 if (arg && ((*arg == 'a') || (*arg == 'A')))
  {
   err_ascii();
   flagascii = 1;
  }
 else if (arg && ((*arg == 'i') || (*arg == 'I') || (*arg == 'l') || (*arg == 'L')))
  {
   /* XXX: pay attention to number of local bits? */
   err_typeok();
   flagascii = 0;
  }
 else
   err_typebad();
}
void ftp_pwd(arg) char *arg;
{
 char *x;
 x = wd_get();
 if (x) out(x);
 else err_pwdio();
}
void ftp_cwd(arg) char *arg;
{
 if (flagwriting) { err_transferring(); return; }
 if (!arg)
   arg = "/"; /* no apparent harm */
 if (wd_set(arg) == -1) err_dirbad();
 else err_wddone();
}
void ftp_cdup(arg) char *arg;
{
 if (flagwriting) { err_transferring(); return; }
 if (wd_set("..") == -1) err_dirbad();
 else err_updone();
}
void ftp_port(arg) char *arg;
{
 int h1; int h2; int h3; int h4; int p1; int p2;
 if (flagwriting) { err_transferring(); return; }
 if (!arg)
   err_noport();
 else
   if (scan6(arg,&h1,&h2,&h3,&h4,&p1,&p2))
    {
     usd_default();
     themd.h1 = h1; themd.h2 = h2; themd.h3 = h3; themd.h4 = h4;
     themd.p1 = p1; themd.p2 = p2;
     passive_off();
     if ((themd.h1 != themc.h1) || (themd.h2 != themc.h2)
	|| (themd.h3 != themc.h3) || (themd.h4 != themc.h4)
	|| (themd.p1 < 4)
	)
      {
       themd = themc;
       err_porthost();
      }
     else
       err_okay();
    }
   else
     err_shortport();
}
void ftp_pasv(arg) char *arg;
{
 if (flagwriting) { err_transferring(); return; }
 if (passive_on() == -1)
   err_local();
 else
  {
   buf[fmt227(buf,usd.h1,usd.h2,usd.h3,usd.h4,usd.p1,usd.p2)] = 0;
   out(buf);
  }
}
void ftp_stat(arg) char *arg;
{
 char *t;
 if (arg) { err_statpath(); return; }
 t = buf;
 t += fmt211(t,usc.h1,usc.h2,usc.h3,usc.h4,usc.p1 * 256 + usc.p2
   ,themc.h1,themc.h2,themc.h3,themc.h4,themc.p1 * 256 + themc.p2
   ,"control");
 if (flagactive || flagaccepted)
   t += fmt211(t,usd.h1,usd.h2,usd.h3,usd.h4,usd.p1 * 256 + usd.p2
     ,themd.h1,themd.h2,themd.h3,themd.h4,themd.p1 * 256 + themd.p2
     ,"data");
 else
   t += fmt211short(t,usd.h1,usd.h2,usd.h3,usd.h4,usd.p1 * 256 + usd.p2
     ,"data listening");
 if (flagwriting)
   if (transferlen)
     t += fmttrans2(t,transferpos,transferlen);
   else
     t += fmttrans3(t,transferpos);
 else
   t += fmttrans4(t);
 *t = 0;
 out(buf);
}
static void ftp_get(fn,type) char *fn; int type;
{
 int len;
 if (flagwriting) { err_transferring(); return; }
 len = get_open(fn,type,flagascii);
 if (len < 0)
   switch(len)
    {
     case -2: err_filebad(); return;
     case -3: err_local(); return;
     case -4: err_fileprot(); return;
     case -5: err_filenotreg(); return;
     case -6: err_filenotdir(); return;
     default: err_strange(); return;
    }
 if (flagactive)
  {
   if (type != GET_FILE) err_tryingdir();
   else if (flagascii) err_tryingascii();
   else err_trying();
   fddata = conn_connect(&usd,&themd);
   if (fddata == -1) { err_conn(errno); get_close(); return; }
  }
 else
   if (type != GET_FILE) err_waitingdir();
   else if (flagascii) err_waitingascii();
   else err_waiting();
 transferlen = len;
 transferpos = 0;
 flagwriting = 1;
 flagcorete = 0;
}
void ftp_retr(arg) char *arg;
{ if (!arg) { err_nofile(); return; } ftp_get(arg,GET_FILE); }
void ftp_list(arg) char *arg;
{ if (!arg) ftp_get(".",GET_EPLF); else ftp_get(arg,GET_EPLFONE); }
void ftp_nlst(arg) char *arg;
{ if (!arg) arg = "."; ftp_get(arg,GET_NAMES); }
void ftp_abor(arg) char *arg;
{
 if (flagwriting)
   endtransfer(2);
 err_aborted();
}
