#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
extern int errno;
#include "sys.h"
#include "get.h"
#include "log.h"
#include "fmt6.h"
#include "fmt.h"
#include "str.h"
#include "byte.h"
#include "direntry.h"

static char *lookingfor;
static int flagopen = 0;
char get_buf[4096];
#define buf get_buf
char get_preascbuf[2048];
#define preascbuf get_preascbuf
static int len = 0;
static int pos = 0;

static int readmode;

/* readmode 0: */
static int fd = -1;
static int flagascii;
/* readmode 1: */
static int flaglong;
static DIR *dir;
static direntry *dp;
static int dplen;
static int dppos;

static struct stat st;
static char stbuf[80];
static int stbuflen;

static void dostbuf(fn)
char *fn;
{
 if (stat(fn,&st) == -1)
   stbuflen = fmt_str(stbuf,"+\t");
 else
  {
   stbuflen = fmteplf(stbuf
     ,(unsigned long) st.st_dev
     ,(unsigned long) st.st_ino
     ,(unsigned long) st.st_mtime
     ,((st.st_mode & S_IFMT) == S_IFDIR)
     ,((st.st_mode & S_IFMT) == S_IFREG) && ((st.st_mode & 0444) == 0444)
     ,(unsigned long) st.st_size
     );
  }
}

static int badname(fn,len)
char *fn;
int len;
{
 int i;
 if (len == 0) return 1;
 if (byte_chr(fn,len,' ') < len) return 1;
 if (byte_chr(fn,len,'~') < len) return 1;
 if (byte_chr(fn,len,'\t') < len) return 1;
 if (byte_chr(fn,len,'\r') < len) return 1;
 if (byte_chr(fn,len,'\n') < len) return 1;
 if (byte_chr(fn,len,'\0') < len) return 1; /* currently can't happen */
 return 0;
}

static int quietname(fn,len)
char *fn;
int len;
{
 if (badname(fn,len)) return 1;
 if (!str_diff(fn,".")) return 1;
 if (!str_diff(fn,"..")) return 1;
 if (lookingfor && str_diff(fn,lookingfor)) return 1;
 return 0;
}

void get_close()
{
 if (flagopen)
  {
   if (readmode == 1) { closedir(dir); dir = 0; }
   if (readmode == 0) { close(fd); fd = -1; }
   flagopen = 0;
  }
}

void get_done(w)
int w;
{
 pos += w;
}

char *get_more(n)
int *n;
{
 if (!flagopen) { *n = -1; errno = EBADF; return 0; }
 if (pos >= len)
  {
   switch(readmode)
    {
     case 2:
       if (dppos >= dplen)
	{
	 if (dplen >= 2) { *n = 0; return 0; }
	 dplen = str_len(lookingfor) + 2;
         if (flaglong)
	  {
           dostbuf(lookingfor);
           dppos = -stbuflen;
	  }
         else
	   dppos = 0;
	}
       len = 0;
       while ((len < sizeof(buf)) && (dppos < dplen))
        {
         if (dppos < 0) buf[len++] = stbuf[stbuflen + dppos];
         else if (dppos == dplen - 2) buf[len++] = '\r';
         else if (dppos == dplen - 1) buf[len++] = '\n';
         else buf[len++] = lookingfor[dppos];
         ++dppos;
        }
       break;
     case 1:
       if (dppos >= dplen)
        {
         do
	  {
           dp = readdir(dir);
	   if (dp) dplen = str_len(dp->d_name) + 2;
	  }
         while (dp && quietname(dp->d_name,dplen - 2));
         if (!dp) { *n = 0; return 0; }
         if (flaglong)
	  {
	   dostbuf(dp->d_name);
	   dppos = -stbuflen;
	  }
         else
	   dppos = 0;
        }
       len = 0;
       while ((len < sizeof(buf)) && (dppos < dplen))
        {
         if (dppos < 0) buf[len++] = stbuf[stbuflen + dppos];
         else if (dppos == dplen - 2) buf[len++] = '\r';
         else if (dppos == dplen - 1) buf[len++] = '\n';
         else buf[len++] = dp->d_name[dppos];
         ++dppos;
        }
       break;
     case 0:
       if (flagascii)
	{
	 int i;
	 char ch;
	 len = sys_read(fd,preascbuf,sizeof(preascbuf));
	 if (len <= 0) { *n = len; return 0; }
	 pos = 0;
	 for (i = 0;i < len;++i)
	  {
	   ch = preascbuf[i];
	   if (ch == '\n') buf[pos++] = '\r';
	   buf[pos++] = ch;
	  }
	 len = pos;
	}
       else
	{
	 len = sys_read(fd,buf,sizeof(buf));
         if (len <= 0) { *n = len; return 0; }
	}
       break;
    }
   pos = 0;
  }
 *n = len - pos;
 return buf + pos;
}

int get_open(fn,type,asc)
char *fn;
int type;
int asc;
{
 get_close();

 if (badname(fn,str_len(fn))) return -2;
 if (stat(fn,&st) == -1) return (errno == EIO) ? -3 : -2;

 if (type != GET_FILE)
  {
   flaglong = (type != GET_NAMES);
   if ((type == GET_EPLFONE) || ((st.st_mode & S_IFMT) == S_IFREG))
    {
     lookingfor = fn;
     readmode = 2;
     dppos = dplen = 0;
     flagopen = 1;
     return 0;
    }
   if ((st.st_mode & S_IFMT) != S_IFDIR) return -6;
   if (str_diff(fn,".")) flaglong = 0;
   lookingfor = 0;
   dir = opendir(fn);
   if (!dir) return (errno == EIO) ? -3 : -2;
   readmode = 1;
   dppos = dplen = 0;
   flagopen = 1;
   return 0;
  }

 readmode = 0;
 if ((st.st_mode & S_IFMT) != S_IFREG) return -5;
 if ((st.st_mode & 0444) != 0444) return -4;
 fd = open(fn,O_RDONLY | O_NDELAY);
 if (fd == -1) return (errno == EIO) ? -3 : -2;
 if (fstat(fd,&st) == -1) { close(fd); return (errno == EIO) ? -3 : -2; }
 if ((st.st_mode & S_IFMT) != S_IFREG) { close(fd); return -5; }
 if ((st.st_mode & 0444) != 0444) { close(fd); return -4; }
 log_retr((unsigned long) st.st_dev,(unsigned long) st.st_ino);
 pos = len = 0;
 flagopen = 1;
 flagascii = asc;
 return st.st_size;
}
